##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Upload view for zope.file objects.

"""

__docformat__ = "reStructuredText"

import re

import zope.contenttype.parse
import zope.component
import zope.event
import zope.file.file
import zope.formlib.form

import zope.schema

from zope.file.i18n import _

from zope.container.interfaces import INameChooser

from zope.mimetype.interfaces import IContentInfo
from zope.mimetype.interfaces import ICodecPreferredCharset
from zope.mimetype.interfaces import ICharsetCodec
from zope.mimetype.interfaces import IContentTypeInterface
from zope.mimetype.interfaces import IContentTypeEncoded
from zope.mimetype.interfaces import IMimeTypeGetter

from zope.mimetype.event import changeContentType

from zope.security.proxy import removeSecurityProxy
from zope.lifecycleevent import ObjectModifiedEvent

_nameFinder = re.compile(r'(.*[\\/:])?(.+)')

def nameFinder(fileob):
    name = getattr(fileob, 'filename', None)
    if name is None:
        return None
    match  = _nameFinder.match(name)
    if match is not None:
        match = match.group(2)
    return match


class Upload(zope.formlib.form.AddForm):

    form_fields = zope.formlib.form.fields(
        zope.schema.Bytes(
            __name__="data",
            title=_("Upload data"),
            description=_("Upload file"),
            ),
        )

    def create(self, data):
        ob = self._create_instance(data)
        f = self.request.form["form.data"]
        updateBlob(ob, f)
        self._name = nameFinder(f)
        return ob

    def add(self, ob):
        if self._name and self.context.nameAllowed():
            nc = INameChooser(self.context.context)
            name = nc.chooseName(self._name, ob)
            self.context.contentName = name

        res = super(Upload, self).add(ob)
        # We need to slam on an interface based on the effective MIME type;
        # start by getting the IContentInfo for ob:
        ci = IContentInfo(ob, None)
        if ci is not None and ci.effectiveMimeType:
            # we might have *something*
            iface = zope.component.queryUtility(
                IContentTypeInterface,
                ci.effectiveMimeType)
            if iface is not None:
                changeContentType(ob, iface)
        return res

    def _create_instance(self, data):
        return zope.file.file.File()


class Reupload(zope.formlib.form.Form):

    form_fields = zope.formlib.form.Fields(
        zope.schema.Bytes(
            __name__="data",
            title=_("Upload data"),
            description=_("Upload file to replace the current data"),
            ),
        )

    @zope.formlib.form.action(_("Edit"))
    def upload(self, action, data):
        context = self.context
        old_codec = None
        if "charset" in context.parameters:
            old_codec = zope.component.queryUtility(
                ICharsetCodec,
                context.parameters["charset"])

        if data.get("data") is not None:
            updateBlob(context, self.request.form["form.data"])

        # update the encoding, but only if the new content type is encoded
        encoded = IContentTypeEncoded.providedBy(context)
        if encoded and "charset" in context.parameters:
            codec = zope.component.queryUtility(
                ICharsetCodec,
                context.parameters["charset"])
            if codec and getattr(old_codec, "name", None) != codec.name:
                # use the preferred charset for the new codec
                new_charset = zope.component.getUtility(
                    ICodecPreferredCharset,
                    codec.name)
                parameters = dict(context.parameters)
                parameters["charset"] = new_charset.name
                context.parameters = parameters
        # these subscribers generally expect an unproxied object.
        zope.event.notify(ObjectModifiedEvent(removeSecurityProxy(context)))


def updateBlob(ob, input):
    # Bypass the widget machinery for now; we'd rather have a blob
    # widget that exposes the interesting information from the
    # upload.
    f = input
    # We need to seek back since the form machinery has already
    # read all the data :-(
    f.seek(0)
    data = f.read()

    contentType = f.headers.get("Content-Type")

    mimeTypeGetter = zope.component.getUtility(IMimeTypeGetter)
    mimeType = mimeTypeGetter(data=data, content_type=contentType,
                              name=nameFinder(f))
    if not mimeType:
        mimeType = "application/octet-stream"

    if contentType:
        _major, _minor, parameters = zope.contenttype.parse.parse(
            contentType)
        if "charset" in parameters:
            parameters["charset"] = parameters["charset"].lower()
        ob.mimeType = mimeType
        ob.parameters = parameters
    else:
        ob.mimeType = mimeType
        ob.parameters = {}

    w = ob.open("w")
    try:
        # if we're security proxied, may not be able to use a `with`
        w.write(data)
    finally:
        w.close()



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Functional tests for zope.file.

"""

__docformat__ = "reStructuredText"

import doctest

from six.moves import urllib_parse as urllib

from zope.app.wsgi.testlayer import http
from zope.app.wsgi.testlayer import FakeResponse
from zope.browser.interfaces import IAdding
from zope.browsermenu.menu import getFirstMenuItem
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile

from zope.container.interfaces import IContainerNamesContainer
from zope.container.interfaces import INameChooser

from zope.interface import implementer

from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserPublisher

from zope.security.proxy import removeSecurityProxy

from zope.traversing.browser.absoluteurl import absoluteURL

import zope.app.wsgi.testlayer
import zope.file
import zope.security.checker
import zope.testbrowser.wsgi


class BrowserLayer(zope.testbrowser.wsgi.TestBrowserLayer,
                   zope.app.wsgi.testlayer.BrowserLayer):
    pass


ZopeFileLayer = BrowserLayer(zope.file)


class _FakeResponse(FakeResponse):
    if str is bytes:
        # py2 has a bug, assuming headers are in unicode already, or
        # are decodable implicitly to ascii
        def getHeaders(self):
            headers = super(_FakeResponse, self).getHeaders()
            result = []
            for key, value in headers:
                assert isinstance(key, str)
                assert isinstance(value, str)
                result.append((key.decode('latin-1'),
                               value.decode('latin-1')))
            return result

        # we also need to ensure we get something that can be printed
        # in ascii and produce the same output as Py3.
        def __str__(self):
            return self.getOutput().decode('latin-1').encode("utf-8")


def FunctionalBlobDocFileSuite(*paths, **kw):
    globs = kw.setdefault('globs', {})
    globs['getRootFolder'] = ZopeFileLayer.getRootFolder
    kw['package'] = doctest._normalize_module(kw.get('package'))

    kwsetUp = kw.get('setUp', lambda x: None)
    def setUp(test):
        wsgi_app = ZopeFileLayer.make_wsgi_app()
        def _http(query_str, *args, **kwargs):
            # Strip leading \n
            query_str = query_str.lstrip()
            response = http(wsgi_app, query_str, *args, **kwargs)
            response.__class__ = _FakeResponse
            return response

        test.globs['http'] = _http
        kwsetUp(test)
    kw['setUp'] = setUp

    if 'optionflags' not in kw:
        old = doctest.set_unittest_reportflags(0)
        doctest.set_unittest_reportflags(old)
        kw['optionflags'] = (old
                             | doctest.ELLIPSIS
                             | doctest.REPORT_NDIFF
                             | doctest.NORMALIZE_WHITESPACE)

    suite = doctest.DocFileSuite(*paths, **kw)
    suite.layer = ZopeFileLayer
    return suite


@implementer(IAdding)
class Adding(BrowserView):
    # set in BrowserView.__init__
    request = None
    context = None
    contentName = None # usually set by Adding traverser

    def add(self, content):
        container = self.context
        name = self.contentName
        chooser = INameChooser(container)

        request = self.request
        name = request.get('add_input_name', name)
        assert name

        chooser.checkName(name, content)

        container[name] = content
        self.contentName = name # Set the added object Name
        return container[name]

    def nextURL(self):
        return absoluteURL(self.context, self.request) + '/@@contents.html'

    def nameAllowed(self):
        """Return whether names can be input by the user."""
        return not IContainerNamesContainer.providedBy(self.context)


class Contents(BrowserView):

    error = ''
    message = ''
    normalButtons = False
    specialButtons = False
    supportsRename = False
    contents = ViewPageTemplateFile('tests/contents.pt')
    contentsMacros = contents

    def listContentInfo(self):
        return self._normalListContentsInfo()

    def _normalListContentsInfo(self):
        items = self.context.items()
        info = [self._extractContentInfo(x) for x in items]
        return info

    def _extractContentInfo(self, item):
        info = {}
        oid, obj = item
        info['id'] = info['cb_id'] = oid
        info['object'] = obj
        info['url'] = urllib.quote(oid.encode('utf-8'))
        return info


@implementer(IBrowserPublisher)
class ManagementViewSelector(BrowserView):
    """View that selects the first available management view.

    Support 'zmi_views' actions like: 'javascript:alert("hello")',
    '../view_on_parent.html' or '++rollover++'.
    """
    # Copied from zope.app.publication
    # Simplified to assert just the test case we expect.

    def browserDefault(self, request):
        return self, ()

    def __call__(self):
        item = getFirstMenuItem('zmi_views', self.context, self.request)
        assert item
        redirect_url = item['action']
        if not redirect_url.lower().startswith(('../', 'javascript:', '++')):
            self.request.response.redirect(redirect_url)
            return u''

        # The zope.app.publication version would redirect to /
        # with self.request.response.redirect('.) and return u''.
        # But our tests never get here.
        raise AssertionError("We shouldn't get here") # pragma: no cover



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""The 'Content Type' view for files.

"""


__docformat__ = "reStructuredText"

from contextlib import closing

import zope.component
import zope.formlib.form
import zope.formlib.interfaces
import zope.interface
import zope.schema

from zope.mimetype.interfaces import IContentTypeEncoded
from zope.mimetype.interfaces import IContentTypeInterface
from zope.mimetype.interfaces import ICharsetCodec
from zope.mimetype.interfaces import ICodecPreferredCharset

from zope.mimetype.event import changeContentType

from zope.mimetype.source import contentTypeSource
from zope.mimetype.source import codecSource

from zope.file.i18n import _

from zope.security.proxy import removeSecurityProxy

from zope.lifecycleevent import ObjectModifiedEvent


def validateCodecUse(file, iface, codec, codec_field):
    """Validate the use of `codec` for the data in `file`.

    `iface` is used as the content-type interface for `file`.  If
    `iface` is None, no validation is performed and no error is
    reported (this validation is considered irrelevant).

    If `codec` is None, no error is reported, since that indicates
    that no codec is known to be accurate for the data in `file`.

    `codec_field` is the form field that is used to select the codec;
    it is used to generate an error should one be necessary.

    Returns a list of widget input errors that should be added to any
    other errors the form determines are relevant.

    """
    if codec is None or iface is None:
        return []
    if not iface.extends(IContentTypeEncoded):
        return []

    errs = []
    # Need to test that this codec can be used for this
    # document:
    with closing(file.open("r")) as f:
        content_data = f.read()

    try:
        _, consumed = codec.decode(content_data)
        if consumed != len(content_data):
            raise UnicodeError("not all data decoded")
    except UnicodeError:
        err = zope.formlib.interfaces.WidgetInputError(
            codec_field.__name__,
            codec_field.field.title,
            "Selected encoding cannot decode document.")
        errs.append(err)
    return errs


class ContentTypeForm(zope.formlib.form.Form):

    mimeType_field = zope.formlib.form.Field(
        zope.schema.Choice(
            __name__="mimeType",
            title=_("Content type"),
            description=_("Type of document"),
            source=contentTypeSource,
            ))

    encoding_field = zope.formlib.form.Field(
        zope.schema.Choice(
            __name__="encoding",
            title=_("Encoding"),
            description=_("Character data encoding"),
            source=codecSource,
            required=False,
            ))

    def get_rendered_encoding(self):
        charset = self.context.parameters.get("charset")
        if charset:
            return zope.component.queryUtility(ICharsetCodec, charset)

    encoding_field.get_rendered = get_rendered_encoding

    def get_rendered_mimeType(self):
        # make sure we return the exact `IContentType`-derived
        # interface that the context provides; there *must* be only
        # one!
        ifaces = zope.interface.directlyProvidedBy(
            removeSecurityProxy(self.context))
        for iface in ifaces:
            if IContentTypeInterface.providedBy(iface):
                return iface

    mimeType_field.get_rendered = get_rendered_mimeType

    def setUpWidgets(self, ignore_request=False):
        # We need to re-compute the fields before initializing the widgets
        fields = [self.mimeType_field]
        if IContentTypeEncoded.providedBy(self.context):
            self.have_encoded = True
            fields.append(self.encoding_field)
        else:
            self.have_encoded = False
        self.form_fields = zope.formlib.form.Fields(*fields)
        super(ContentTypeForm, self).setUpWidgets(
            ignore_request=ignore_request)

    def save_validator(self, action, data):
        errs = self.validate(None, data)
        errs += validateCodecUse(
            self.context, data.get("mimeType"), data.get("encoding"),
            self.encoding_field)
        return errs

    @zope.formlib.form.action(_("Save"), validator=save_validator)
    def save(self, action, data):
        context = self.context
        if data.get("mimeType"):
            #
            # XXX Note that the content type parameters are not
            # modified; ideally, the IContentInfo adapter would filter
            # the parameters for what makes sense for the current
            # content type.
            #
            iface = data["mimeType"]
            unwrapped = removeSecurityProxy(context)
            changeContentType(unwrapped, iface)

        # update the encoding, but only if the new content type is encoded
        encoded = IContentTypeEncoded.providedBy(context)
        # We only care about encoding if we're encoded now and were also
        # encoded before starting the re
        if encoded and self.have_encoded:
            codec = data["encoding"]
            if codec is None:
                # remove any charset found, since the user said it was wrong;
                # what it means to have no known charset is that the policy
                # (IContentInfo) will have to decide how to treat encoding.
                if "charset" in context.parameters:
                    # We can't just "del" the existing value, since
                    # the security model does not like that.  We can
                    # set a new value for context.parameters, though.
                    parameters = dict(context.parameters)
                    del parameters["charset"]
                    context.parameters = parameters
            else:
                if "charset" in context.parameters:
                    old_codec = zope.component.queryUtility(
                        ICharsetCodec,
                        context.parameters["charset"])
                else:
                    old_codec = None
                if getattr(old_codec, "name", None) != codec.name:
                    # use the preferred charset for the new codec
                    new_charset = zope.component.getUtility(
                        ICodecPreferredCharset,
                        codec.name)
                    parameters = dict(context.parameters)
                    parameters["charset"] = new_charset.name
                    context.parameters = parameters

        zope.event.notify(
            ObjectModifiedEvent(removeSecurityProxy(context)))





import unittest

from zope import component
from zope.interface import implementer

from zope.mimetype.interfaces import IContentTypeEncoded
from zope.mimetype.interfaces import ICharsetCodec
from zope.mimetype.interfaces import ICodecPreferredCharset
from zope.mimetype.interfaces import IMimeTypeGetter

from zope.mimetype.typegetter import smartMimeTypeGuesser

from io import BytesIO

from zope.file.tests import skipWithoutZopeFormlib

@implementer(IContentTypeEncoded)
class MockContext(object):
    def __init__(self):
        self.parameters = {}

    def open(self, _):
        return self

    def write(self, _):
        pass

    def close(self):
        pass

@skipWithoutZopeFormlib
class TestFunctions(unittest.TestCase):

    def test_name_finder_with_no_filename(self):
        from zope.file import upload
        class MockFile(object):
            pass

        res = upload.nameFinder(MockFile())

        self.assertIsNone(res)

@skipWithoutZopeFormlib
class TestUpdateBlob(unittest.TestCase):

    def setUp(self):
        super(TestUpdateBlob, self).setUp()
        component.provideUtility(smartMimeTypeGuesser, IMimeTypeGetter)
        self.context = MockContext()
        self.data = BytesIO()
        self.data.headers = {}

    def tearDown(self):
        gsm = component.getGlobalSiteManager()
        gsm.unregisterUtility(smartMimeTypeGuesser, IMimeTypeGetter)
        super(TestUpdateBlob, self).tearDown()

    def test_no_mime_type(self):
        from zope.file import upload
        upload.updateBlob(self.context, self.data)
        self.assertEqual(self.context.mimeType, 'application/octet-stream')
        self.assertEqual(self.context.parameters, {})

@skipWithoutZopeFormlib
class TestReupload(unittest.TestCase):

    def test_different_codec(self):
        from zope.file import upload

        class MockCodec(object):
            name = 'TestContentTypeForm'

        class MockRequest(object):
            form = ()
            headers = ()

        context = MockContext()
        orig_codec_name = context.parameters['charset'] = 'Original'

        request = MockRequest()
        form = upload.Reupload(context, request)

        sio = BytesIO()
        request.form = {'form.data': sio}
        sio.headers = {'Content-Type': 'text/plain; charset=TestContentTypeForm'}

        form.have_encoded = True
        data = {}
        codec = data['encoding'] = MockCodec()

        data['data'] = 1

        orig_codec = MockCodec()
        orig_codec.name = orig_codec_name

        component.provideUtility(orig_codec, ICharsetCodec, orig_codec_name)
        component.provideUtility(codec, ICharsetCodec, codec.name.lower())
        component.provideUtility(codec, ICodecPreferredCharset, codec.name)
        component.provideUtility(smartMimeTypeGuesser, IMimeTypeGetter)
        try:
            form.upload.success_handler(form, "action", data)
            self.assertIn('charset', context.parameters)
            self.assertEqual(codec.name, context.parameters['charset'])
        finally:
            gsm = component.getGlobalSiteManager()
            gsm.unregisterUtility(codec,
                                  ICodecPreferredCharset, codec.name.lower())
            gsm.unregisterUtility(orig_codec, ICharsetCodec, orig_codec_name)
            gsm.unregisterUtility(smartMimeTypeGuesser, IMimeTypeGetter)

def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Interfaces for zope.file.

These interfaces support efficient access to file data.

"""
__docformat__ = "reStructuredText"

import zope.interface
import zope.mimetype.interfaces
import zope.schema

from zope.file.i18n import _


class IFile(zope.mimetype.interfaces.IContentTypeAware):
    """File content object for Zope."""

    def open(mode="r"):
        """Return an object providing access to the file data.

        Allowed values for `mode` are 'r' (read); 'w' (write); 'a' (append) and
        'r+' (read/write).  Other values cause `ValueError` to be raised.

        If the file is opened in read mode, an object with an API (but
        not necessarily interface) of `IFileReader` is returned; if
        opened in write mode, an object with an API of `IFileWriter` is
        returned; if in read/write, an object that implements both is
        returned.

        All readers and writers operate in 'binary' mode.

        """

    def openDetached():
        """Return file data disconnected from database connection.

        Read access only.
        """

    size = zope.schema.Int(
        title=_("Size"),
        description=_("Size in bytes"),
        readonly=True,
        required=True,
        )


# The remaining interfaces below serve only to document the kind of APIs
# to be expected, as described in IFile.open above.

class IFileAccessor(zope.interface.Interface):
    """Base accessor for `IFileReader` and `IFileWriter`."""

    def close():
        """Release the accessor.

        Resources held by the accessor are freed, and further use of
        the accessor will result in a `ValueError` exception.

        """

    mode = zope.schema.BytesLine(
        title=_("Open mode passed to the corresponding file's open() method,"
                " or the default value if not passed."),
        required=True,
        )


class IFileReader(IFileAccessor):
    """Read-accessor for file objects.

    The methods here mirror the corresponding portions of the API for
    Python file objects.

    """

    def read(size=-1):
        """Read at most `size` bytes from the reader.

        If the size argument is negative or omitted, read all data
        until EOF is reached.  The bytes are returned as a str.  An
        empty string is returned when EOF is encountered immediately.

        """

    def seek(offset, whence=0):
        """Set the reader's current position.

        The `offset` is a number of bytes.  The direction and source
        of the measurement is determined by the `whence` argument.

        The `whence` argument is optional and defaults to 0 (absolute
        file positioning); other recognized values are 1 (seek
        relative to the current position) and 2 (seek relative to the
        end).

        There is no return value.

        """

    def tell():
        """Return the current position, measured in bytes."""


class IFileWriter(IFileAccessor):
    """Write-accessor for file objects.

    The methods here mirror the corresponding portions of the API for
    Python file objects.

    """

    def flush():
        """Ensure the file object has been updated.

        This is used to make sure that any data buffered in the
        accessor is stored in the file object, allowing other
        accessors to use it.

        """

    def write(bytes):
        """Write a string to the file.

        Due to buffering, the string may not actually show up in the
        file until the `flush()` or `close()` method is called.

        There is no return value.

        """



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from zope import interface, component
import zope.filerepresentation.interfaces
from zope.file import interfaces

@component.adapter(interfaces.IFile)
@interface.implementer(zope.filerepresentation.interfaces.IReadFile)
class ReadFileAdapter(object):

    def __init__(self, context):
        self.context = context

    def size(self):
        return self.context.size

    def read(self):
        with self.context.open() as f:
            return f.read()

@component.adapter(interfaces.IFile)
@interface.implementer(zope.filerepresentation.interfaces.IWriteFile)
class WriteFileAdapter(object):

    def __init__(self, context):
        self.context = context

    def write(self, data):
        with self.context.open('w') as f:
            f.write(data)



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Event support code.

"""
__docformat__ = "reStructuredText"


def updateMimeType(file, event):
    """Update the file's mimeType on a change.

    If the current mimeType value is known to be legal for the new
    content type, leave it alone.  If not, set the mimeType attribute
    to the first value from the list of known MIME types for the
    content type.

    """
    if event.newContentType is not None:
        types = event.newContentType.getTaggedValue("mimeTypes")
        if file.mimeType not in types:
            file.mimeType = types[0]



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for zope.file.

"""


__docformat__ = "reStructuredText"
import doctest
import unittest
import re

import zope.file
from zope.component.testlayer import ZCMLFileLayer
from zope.testing import renormalizing

checker = renormalizing.RENormalizing([
    # Python 3 unicode removed the "u".
    (re.compile("u('.*?')"), r"\1"),
    (re.compile('u(".*?")'), r"\1"),
    # Python 3 bytes added the "b".
    (re.compile("b('.*?')"), r"\1"),
    (re.compile('b(".*?")'), r"\1"),
])


ZopeFileLayer = ZCMLFileLayer(zope.file, 'configure.zcml')

def fromDocFile(path, **kwargs):
    try:
        from zope.file import testing
    except ImportError:
        def setUp(test):
            raise unittest.SkipTest("Unable to import zope.file.testing")
        suite = doctest.DocFileSuite(path, setUp=setUp, **kwargs)
    else:
        suite = testing.FunctionalBlobDocFileSuite(path, **kwargs)
    return suite



def fromSimpleDocFile(path, **kwargs):
    suite = doctest.DocFileSuite(path, checker=checker)
    suite.layer = ZopeFileLayer
    return suite

def test_suite():
    return unittest.TestSuite((
        doctest.DocFileSuite("../README.rst", checker=checker),
        doctest.DocFileSuite("../browser.rst", checker=checker),
        fromSimpleDocFile("../adapters.rst", checker=checker),
        fromDocFile('../contenttype.rst', checker=checker),
        fromDocFile("../download.rst", checker=checker),
        fromDocFile("../upload.rst", checker=checker),
        ))



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Implementation of the file content type.

"""


from contextlib import closing
import persistent

import zope.location.interfaces
import zope.file.interfaces
import zope.interface

from ZODB.blob import Blob

@zope.interface.implementer(
    zope.file.interfaces.IFile,
    zope.location.interfaces.ILocation)
class File(persistent.Persistent):

    __name__ = None
    __parent__ = None
    mimeType = None

    _data = b""
    size = 0

    def __init__(self, mimeType=None, parameters=None):
        self.mimeType = mimeType
        if parameters is None:
            parameters = {}
        else:
            parameters = dict(parameters)
        self.parameters = parameters
        blob = self._data = Blob()
        with blob.open('w') as fp:
            fp.write(b'')

    # Note that if we are security proxied, we may not be able to use
    # __exit__ on the return values from open*, meaning they can't directly
    # be used in a with statement.

    def open(self, mode="r"):
        return self._data.open(mode)

    def openDetached(self):
        return open(self._data.committed(), 'rb')

    @property
    def size(self):
        if self._data == b"": # pragma: no cover
            # It shouldn't be possible to get here; perhaps this is
            # compatibility code for old objects in existing databases?
            return 0

        with closing(self.open()) as reader:
            reader.seek(0,2)
            size = int(reader.tell())

        return size



import unittest

def skipWithoutZopeFormlib(test):
    try:
        from zope import formlib
    except ImportError:
        formlib = None
    return unittest.skipIf(formlib is None, "Need zope.formlib")(test)



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Presentation adapters for zope.file.

"""
__docformat__ = "reStructuredText"

from zope import interface
from zope import component

import zope.size
import zope.size.interfaces
import zope.file.interfaces

@interface.implementer(zope.size.interfaces.ISized)
@component.adapter(zope.file.interfaces.IFile)
class Sized(object):

    def __init__(self, context):
        self.context = context

    def sizeForSorting(self):
        return "byte", self.context.size

    def sizeForDisplay(self):
        return zope.size.byteDisplay(self.context.size)



##############################################################################
#
# Copyright (c) 2006-2009 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup for zope.file package

"""
import os
from setuptools import setup, find_packages

def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as f:
        return f.read()

version = '1.2.0'

BROWSER_REQUIRES = [
    'zope.browser',
    'zope.browserresource',
    'zope.publisher >= 4.3.1',
    'zope.formlib',
]

TESTS_REQUIRE = BROWSER_REQUIRES + [
    'zope.app.basicskin >= 4.0.0',
    'zope.app.http',
    'zope.app.pagetemplate >= 4.0.0',
    'zope.app.principalannotation',
    'zope.app.publication',
    'zope.app.wsgi',

    'zope.applicationcontrol',
    'zope.copypastemove',
    'zope.browser',
    'zope.browsermenu',
    'zope.login',
    'zope.password',
    'zope.proxy >= 4.2.1',
    'zope.principalregistry',
    'zope.securitypolicy',
    'zope.testbrowser >= 5.2',
    'zope.testrunner',
]

setup(
    name='zope.file',
    version=version,
    author='Zope Corporation and Contributors',
    author_email='zope-dev@zope.org',
    description='Efficient File Implementation for Zope Applications',
    long_description=(
        read('README.rst')
        + '\n\n' +
        '.. contents::'
        + '\n\n' +
        read('src', 'zope', 'file', 'README.rst')
        + '\n\n' +
        read('src', 'zope', 'file', 'download.rst')
        + '\n\n' +
        read('src', 'zope', 'file', 'upload.rst')
        + '\n\n' +
        read('src', 'zope', 'file', 'contenttype.rst')
        + '\n\n' +
        read('src', 'zope', 'file', 'browser.rst')
        + '\n\n' +
        read('CHANGES.rst')
    ),
    keywords="zope3 web html ui file pattern",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: Zope :: 3',
    ],
    url='https://github.com/zopefoundation/zope.file',
    license='ZPL 2.1',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['zope'],
    extras_require={
        'test': TESTS_REQUIRE,
        'browser': BROWSER_REQUIRES,
    },
    install_requires=[
        'setuptools',
        'ZODB',
        'zope.annotation',
        'zope.component',
        'zope.container',
        'zope.contenttype',
        'zope.event',
        'zope.filerepresentation',
        'zope.i18nmessageid',
        'zope.lifecycleevent',
        'zope.interface',
        'zope.location',
        'zope.mimetype >= 2.3.0',
        'zope.schema',
        'zope.security >= 4.1.0',
        'zope.size',
    ],
    include_package_data=True,
    zip_safe=False,
)





import unittest

from zope import component
from zope.interface import implementer

from zope.mimetype.interfaces import IContentTypeEncoded
from zope.mimetype.interfaces import ICodecPreferredCharset

from zope.file.tests import skipWithoutZopeFormlib

@skipWithoutZopeFormlib
class TestCodec(unittest.TestCase):

    def test_unicode_error_too_short(self):
        from zope.file.contenttype import ContentTypeForm
        from zope.file.contenttype import validateCodecUse
        class MockFile(object):
            def open(self, _mode):
                return self

            def read(self):
                return b"Some Data"

            def close(self):
                pass

        class MockCodec(object):
            def decode(self, data):
                return '', 0

        class MockIface(IContentTypeEncoded):
            pass


        errs = validateCodecUse(MockFile(), MockIface,
                                MockCodec(), ContentTypeForm.encoding_field)

        self.assertEqual(1, len(errs))
        self.assertEqual("Selected encoding cannot decode document.",
                         errs[0].errors)

@skipWithoutZopeFormlib
class TestContentTypeForm(unittest.TestCase):

    def test_del_charset_no_encoding(self):
        from zope.file.contenttype import ContentTypeForm
        @implementer(IContentTypeEncoded)
        class MockContext(object):
            def __init__(self):
                self.parameters = {}

        context = MockContext()
        context.parameters['charset'] = "INVALID"

        form = ContentTypeForm(context, None)

        form.have_encoded = True
        data = {}
        data['encoding'] = None

        form.save.success_handler(form, "action", data)
        self.assertNotIn('charset', context.parameters)

    def test_charset_with_encoding(self):
        from zope.file.contenttype import ContentTypeForm
        @implementer(IContentTypeEncoded)
        class MockContext(object):
            def __init__(self):
                self.parameters = {}

        class MockCodec(object):
            name = 'TestContentTypeForm'

        context = MockContext()
        context.parameters['charset'] = "INVALID"

        form = ContentTypeForm(context, None)

        form.have_encoded = True
        data = {}
        codec = data['encoding'] = MockCodec()

        component.provideUtility(codec, ICodecPreferredCharset, codec.name)
        try:
            form.save.success_handler(form, "action", data)
            self.assertIn('charset', context.parameters)
            self.assertEqual(codec.name, context.parameters['charset'])
        finally:
            component.getGlobalSiteManager().unregisterUtility(codec,
                                                               ICodecPreferredCharset, codec.name)

def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



##############################################################################
#
# Copyright (c) 2005 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Download view for files.

"""
__docformat__ = "reStructuredText"

import zope.interface
import zope.mimetype.interfaces
import zope.publisher.browser
import zope.publisher.interfaces.http
import zope.security.proxy

class Download(zope.publisher.browser.BrowserView):

    def __call__(self):
        for k, v in getHeaders(self.context, contentDisposition="attachment"):
            self.request.response.setHeader(k, v)
        return DownloadResult(self.context)


class Inline(zope.publisher.browser.BrowserView):

    def __call__(self):
        for k, v in getHeaders(self.context, contentDisposition="inline"):
            self.request.response.setHeader(k, v)
        return DownloadResult(self.context)


class Display(zope.publisher.browser.BrowserView):

    def __call__(self):
        for k, v in getHeaders(self.context):
            self.request.response.setHeader(k, v)
        return DownloadResult(self.context)

def getHeaders(context, contentType=None, downloadName=None,
               contentDisposition=None, contentLength=None):
        if not contentType:
            cti = zope.mimetype.interfaces.IContentInfo(context, None)
            if cti is not None:
                contentType = cti.contentType
        contentType = contentType or "application/octet-stream"
        headers = ("Content-Type", contentType),

        downloadName = downloadName or context.__name__
        if contentDisposition:
            if downloadName:
                # Headers must be native strings. Under Py2, we probably
                # need to encode the name
                if str is bytes:
                    if not isinstance(downloadName, bytes):
                        downloadName = downloadName.encode("utf-8")
                else:
                    # Under Python 3, we need to pass the native string,
                    # but characters greater than 256 will fail to be encoded
                    # by the WSGI server, which should be using latin-1. The solution
                    # is to smuggle them through using double encoding, allowing the
                    # final recipient to decode its string using utf-8
                    downloadName = downloadName.encode('utf-8').decode('latin-1')

                contentDisposition += (
                    '; filename="%s"' % downloadName
                    )
            headers += ("Content-Disposition", contentDisposition),

        if contentLength is None:
            contentLength = context.size
        headers += ("Content-Length", str(contentLength)),
        return headers


@zope.interface.implementer(zope.publisher.interfaces.http.IResult)
class DownloadResult(object):
    """Result object for a download request."""

    def __init__(self, context):
        self._iter = bodyIterator(
            zope.security.proxy.removeSecurityProxy(context.openDetached()))

    def __iter__(self):
        return self._iter

CHUNK_SIZE = 64 * 1024


def bodyIterator(f):
    try:
        while True:
            chunk = f.read(CHUNK_SIZE)
            if not chunk:
                return
            yield chunk
    finally:
        f.close()
